;; Generated package description from autorevert-tail-truncate.el  -*- no-byte-compile: t -*-
(define-package "autorevert-tail-truncate" "1.0pre0.20250207.125358" "auto-revert-tail your files with RAM control" '((emacs "29.1")) :commit "d1b9c3f6fb8ba963d562a77539ea50ab34e0c988" :authors '(("Stephane Marks" . "shipmints@gmail.com")) :maintainer '("Stephane Marks" . "shipmints@gmail.com") :keywords '("convenience" "tools" "log files" "autorevert") :url "https://github.com/shipmints/autorevert-tail-truncate.el")
